import logging
from rest_framework import serializers
from django.utils.translation import gettext_lazy as _
from taggit_serializer.serializers import TaggitSerializer, TagListSerializerField
from fiction_outlines.models import Series, Character, Location, Outline
from fiction_outlines.models import CharacterInstance, LocationInstance, Arc
from fiction_outlines.models import ArcElementNode, StoryElementNode, MACE_TYPES


logger = logging.getLogger('fiction-outlines-api')

MOVE_POSITION_CHOICES = (
    ('first-child', 'first-child'),
    ('last-child', 'last-child'),
    ('first-sibling', 'first-sibling'),
    ('last-sibling', 'last-sibling'),
    ('left', 'left'),
    ('right', 'right'),
)


def convert_annotated_list(annotated_list, serializer_class):
    '''
    Takes an annotated list and a serializer class and returns a version that is suitable for
    serialization.
    '''
    converted_list = []  # pragma: no cover
    for item, info in annotated_list:
        converted_list.append((serializer_class(item).data, info))
    return converted_list


class SeriesSerializer(TaggitSerializer, serializers.ModelSerializer):
    '''
    Serializer for Series model.
    '''

    tags = TagListSerializerField()

    class Meta:  # pragma: no cover
        model = Series
        fields = ('id', 'title', 'description', 'tags')
        read_only_fields = ('id',)


class CharacterInstanceSerializer(serializers.ModelSerializer):
    '''
    Serializer for character instance.
    '''
    name = serializers.CharField(read_only=True, source='character.name')

    class Meta:  # pragma: no cover
        model = CharacterInstance
        fields = ('id', 'outline', 'character', 'name', 'main_character', 'pov_character', 'protagonist',
                  'antagonist', 'obstacle', 'villain')
        read_only_fields = ('id', 'outline', 'character', 'name')
        extra_kwargs = {}
        for field in read_only_fields:
            extra_kwargs['field'] = {'required': False}


class LocationInstanceSerializer(serializers.ModelSerializer):
    '''
    Serializer for location instance.
    '''
    name = serializers.CharField(read_only=True, source='location.name')

    class Meta:  # pragma: no cover
        model = LocationInstance
        fields = ('id', 'location', 'name', 'outline')
        read_only_fields = ('id', 'location', 'name', 'outline')
        extra_kwargs = {}
        for field in read_only_fields:
            extra_kwargs['field'] = {'required': False}


class CharacterSerializer(TaggitSerializer, serializers.ModelSerializer):
    '''
    Serializer for Character model.
    '''

    tags = TagListSerializerField()
    series = serializers.PrimaryKeyRelatedField(many=True, allow_null=True, queryset=Series.objects.all())
    character_instances = CharacterInstanceSerializer(many=True, read_only=True,
                                                      required=False, source='characterinstance_set')

    class Meta:  # pragma: no cover
        model = Character
        fields = ('id', 'name', 'description', 'series', 'tags', 'character_instances')
        read_only_fields = ('id', 'character_instances')


class LocationSerializer(TaggitSerializer, serializers.ModelSerializer):
    '''
    Serializer for Location model.
    '''

    tags = TagListSerializerField()
    series = serializers.PrimaryKeyRelatedField(many=True, allow_null=True, queryset=Series.objects.all())
    location_instances = LocationInstanceSerializer(many=True, read_only=True,
                                                    required=False, source='locationinstance_set')

    class Meta:  # pragma: no cover
        model = Location
        fields = ('id', 'name', 'description', 'series', 'tags', 'location_instances')
        read_only_fields = ('id', 'location_instances')


class OutlineSerializer(TaggitSerializer, serializers.ModelSerializer):
    '''
    Serializer for Outline model.
    Also provides ``outline_structure``, which is a treebeard annotated list of story element nodes.

    Special read-only fields:

    :attribute length_estimate:
        Auto calculated length (in words) based on the number of characters, locations, and arcs.
    :attribute arc_set:
        The :class:`fiction_outlines.models.Arc` objects associated with this outline.
    :attribute characterinstance_set:
        Character instances associated with this outline.
    :attribute locationinstance_set:
        Location instances associated with this outline.
    :attribute outline_structure:
        An annotated list of :class:`fiction_outlines.models.StoryElementNode` objects for visualizing the
        tree structure.
    '''

    length_estimate = serializers.IntegerField(read_only=True,
                                               required=False,
                                               help_text=_("Autogenerated length estimate for the story in number of words."))  # noqa: E501
    tags = TagListSerializerField()
    series = serializers.PrimaryKeyRelatedField(allow_null=True, queryset=Series.objects.all())
    arc_set = serializers.PrimaryKeyRelatedField(many=True, required=False, read_only=True)
    characterinstance_set = serializers.PrimaryKeyRelatedField(many=True, required=False, read_only=True)
    locationinstance_set = serializers.PrimaryKeyRelatedField(many=True, required=False, read_only=True)
    outline_structure = serializers.SerializerMethodField(read_only=True,
                                                          help_text=_("Annotated list of outline items."))

    def get_outline_structure(self, obj):
        '''
        Retrieves all :class:`fiction_outlines.StoryElementNode` objects associated with this outline,
        and returns it as an annotated list that can be serialized.

        :param obj:
            The outline object this serializer represents.

        :returns: An annotated list for serialization or an empty list.
        '''
        logger.debug("Checking outline structure.")
        if obj.story_tree_root:
            logger.debug('Found a root node')
            annotated_list = StoryElementNode.get_annotated_list(obj.story_tree_root)
            logger.debug('Annotated list retrieved.')
            if len(annotated_list) > 1:
                # There is more than just a single root node: serialize and return.
                annotated_list = convert_annotated_list(annotated_list, StoryElementNodeSerializer)
                logger.debug("Returning results to field.")  # pragma: no cover
                return annotated_list[1:]
        return []

    class Meta:  # pragma: no cover
        model = Outline
        fields = ('id', 'title', 'description', 'series', 'tags', 'length_estimate', 'arc_set',
                  'characterinstance_set', 'locationinstance_set', 'outline_structure')
        read_only_fields = ('id', 'length_estimate', 'arc_set', 'characterinstance_set',
                            'locationinstance_set', 'outline_structure')


class ArcSerializer(serializers.ModelSerializer):
    '''
    Serializer for Arc model.

    To create an Arc, make sure to instead use the :class:`ArcCreateSerializer` and
    post it to :class:`fiction_outlines_api.views.ArcCreateView`
    '''

    outline = serializers.PrimaryKeyRelatedField(read_only=True, required=False)
    current_errors = serializers.ReadOnlyField(read_only=True, required=False,
                                               help_text=_('dict current validation errors'))
    arc_structure = serializers.SerializerMethodField(read_only=True, help_text=_('Annotated list of arc elements.'))

    def get_arc_structure(self, obj):
        '''
        Fetches all the :class:`fiction_outlines.models.ArcElementNode` objects assocaited with this arc as an
        annotated list.

        :param obj:
            The arc this serializer represents.

        :returns: An annotated list ready to be serialized or an empty list.
        '''
        logger.debug("Attempting to retrieve arc structure for arc %s" % obj.pk)
        if obj.arc_root_node:
            annotated_list = ArcElementNode.get_annotated_list(obj.arc_root_node)
            logger.debug("Retrieved annotated list")
            if len(annotated_list) > 1:
                # There is more than just a root node which we don't want to display to users.
                logger.debug("Arc structure found, builidng serialized structure.")
                annotated_list = convert_annotated_list(annotated_list, ArcElementNodeSerializer)
                logger.debug("Success serializing arc!")
                logger.debug("Arc looks like this:\n\n%s" % annotated_list[1:])
                logger.debug("Returning result to field.")
                return annotated_list[1:]
            else:
                logger.debug("Only a root node... returning empty list.")
        return []  # pragma: no cover This is really unlikely to occur.

    class Meta:  # pragma: no cover
        model = Arc
        fields = ('id', 'name', 'mace_type', 'outline', 'name', 'current_errors', 'arc_structure')
        read_only_fields = ('id', 'outline', 'current_errors', 'arc_structure')


class ArcCreateSerializer(serializers.Serializer):
    '''
    Serializer used for arc creation. You can't create an Arc using it's primary serializer.
    '''
    mace_type = serializers.ChoiceField(choices=MACE_TYPES)
    name = serializers.CharField(max_length=255)


class ArcElementNodeSerializer(serializers.ModelSerializer):
    '''
    Serializer for ArcElementNode.

    Special read-only attributes:

    :attribute is_milestone:
        Does this element represent a milestone in the :class:`Arc`
    :attribute milestone_seq:
        If a milestone, what is the proscribed sequence where it should appear
        in the context of the other milestones.
    :attribute parent_outline:
        Convenience reference to the parent outline object.
    :attribute headline:
        Automatically generated from the first line of the description.
    '''

    arc = serializers.PrimaryKeyRelatedField(read_only=True, required=False)
    headline = serializers.ReadOnlyField(required=False, help_text=_('Auto generated from description'))
    story_element_node = serializers.PrimaryKeyRelatedField(required=False, allow_null=True,
                                                            queryset=StoryElementNode.objects.all())
    parent_outline = serializers.PrimaryKeyRelatedField(source='arc.outline', read_only=True, required=False)
    milestone_seq = serializers.IntegerField(allow_null=True, read_only=True, required=False,
                                             help_text=_('If a milestone, what order *should* it appear in?'))

    class Meta:  # pragma: no cover
        model = ArcElementNode
        fields = ('id', 'arc', 'arc_element_type', 'headline', 'description', 'story_element_node',
                  'assoc_characters', 'assoc_locations', 'milestone_seq', 'is_milestone', 'parent_outline')
        read_only_fields = ('id', 'arc', 'headline', 'milestone_seq', 'is_milestone', 'parent_outline')


class StoryElementNodeSerializer(serializers.ModelSerializer):
    '''
    Serializer for StoryElementNode

    Two special fields here:

    :attribute all_characters:
        All characters associated with this node or its descendants.
    :attribute all_locations:
        All locations associated with this node of its descendants.
    '''

    all_characters = CharacterInstanceSerializer(many=True, read_only=True, required=False,
                                                 help_text=_('All characters associated with this node its descendants.'))  # noqa: E501
    all_locations = LocationInstanceSerializer(many=True, read_only=True, required=False,
                                               help_text=_('All locations associated with this node or its descendants.'))  # noqa: E501
    outline = serializers.PrimaryKeyRelatedField(read_only=True, required=False)

    class Meta:  # pragma: no cover
        model = StoryElementNode
        fields = ('id', 'story_element_type', 'name', 'description', 'outline', 'assoc_characters',
                  'assoc_locations', 'all_characters', 'all_locations', 'impact_rating')
        read_only_fields = ('all_characters', 'all_locations', 'impact_rating', 'outline', 'id')



from rest_framework.exceptions import APIException
from django.utils.translation import gettext_lazy as _


class TreeUnavailable(APIException):
    status_code = 503
    default_detail = _('We are having temporary issues with outlines at this time. Please try again later.')
    default_code = 'tree_unavailable'


class ServiceUnavailable(APIException):
    status_code = 503
    default_detail = _('Service temporarily unavailable, please try again later.')
    default_code = 'service_unavailable'



import logging
from django.db import transaction
from django.utils.translation import gettext_lazy as _
from django.http import Http404
from treebeard.exceptions import InvalidPosition, InvalidMoveToDescendant, PathOverflow
from rest_framework.generics import get_object_or_404
from rest_framework.exceptions import ValidationError, PermissionDenied
from rest_framework import response, status
from fiction_outlines.signals import tree_manipulation
from fiction_outlines.models import IntegrityError
from .exceptions import TreeUnavailable
from fiction_outlines.models import ArcGenerationError


POSITIONS = ('first-child', 'last-child', 'first-sibling', 'last-sibling', 'left', 'right')

logger = logging.getLogger('fiction-outlines-api')


class MultiObjectPermissionsMixin(object):
    '''
    API Mixin that compares ``n`` objects and their permissions and returns if both are valid.
    Objects can be referred to by key afterwards in self.object_dict.
    permission dict should be:

    {'obj1': {'obj_class': class_model, 'lookup_url_kwarg', 'object_permission_required': 'your rules perm here'}, ...}

    Is it ugly, yes. Does it make doing this in view after view repeatable. Yep.
    '''
    object_class_permission_dict = {}

    def get_permission_object(self, pkval, objectpermission, obj_class):
        '''
        Fetches the permission object and checks it agains the permission specified.

        :param pkval:
            The pk value to use in the search.
        :param objectpermission:
            A permission based on the :module:`rules` format, e.g. "app_name.permission_name"
        :param obj_class:
            The class of the object we are searching for.

        :returns: The found object.
        '''
        queryset = self.filter_queryset(self.get_object_permission_queryset(obj_class))
        filter_kwargs = {self.lookup_field: pkval}
        obj = get_object_or_404(queryset, **filter_kwargs)
        self.check_object_permissions(self.request, objectpermission, obj)
        return obj

    def check_object_permissions(self, request, perm, obj):
        '''
        Verifies the object permissions are valid for the current user.

        :param request:
             The request object. The user is retrieved from here.
        :param perm:
             The permission to validate against (in :module:`rules`) format.
        :param obj:
             The object to evaluate.

        :raise PermissionDenied: when user lacks the required permission.
        '''
        if not request.user.has_perm(perm, obj):
            raise PermissionDenied
        return super().check_object_permissions(request, obj)

    def post(self, request, *args, **kwargs):
        self.permission_object_dict = {}
        for obj, attrs in self.object_class_permission_dict.items():
            try:
                self.permission_object_dict[obj] = self.get_permission_object(kwargs[attrs['lookup_url_kwarg']],
                                                                              attrs['object_permission_required'],
                                                                              attrs['obj_class'])
            except PermissionDenied as PD:
                error_response = response.Response({'error_message': str(PD)}, status=status.HTTP_403_FORBIDDEN,
                                                   content_type='application/json')
                error_response.accepted_renderer = request.accepted_renderer
                return error_response
        return super().post(request, *args, **kwargs)

    def get_object_permission_queryset(self, obj_class):
        '''
        Fetches the queryset for the permission object.

        :param obj_class:
            Model class of the object to retrieve.

        :returns: A django queryset for the class.
        '''
        return obj_class.objects.all()


class NodeAddMixin(object):
    '''
    API Mixin for add_sibling and add_child commands.

    :attribute fields_required_for_add:
        A tuple of fields that should be required in the submitted serializer in order to create the object.
    '''
    fields_required_for_add = ('description',)  # Declare in more detail in subclass.

    def post(self, request, *args, **kwargs):
        '''
        Parses kwargs and kicks off evaluation process.

        :keyword action:
            Either ``add_child`` or ``add_sibling``.
        :keyword position:
            If ``add_sibling`` this is required, and must be a value from :data:POSITIONS.
            If action is ``add_child`` this is ignored.

        :raise Http404: if action is not a one of the two permitted options.
        '''
        self.source_node = self.get_object()
        self.action = kwargs['action']
        if self.action not in ['add_child', 'add_sibling']:
            raise Http404
        logger.debug('Action is set as %s' % self.action)
        if 'position' in kwargs.keys() and self.action == 'add_sibling':
            self.pos = kwargs['position']
            if self.pos not in POSITIONS:
                raise ValidationError(_("Not a valid position for adding node."))
        else:
            self.pos = None
        return super().post(request, *args, **kwargs)

    def create(self, request, *args, **kwargs):
        '''
        Does the final validation of submitted data and if valid proceeds with creation.

        :raises NotImplementedError: if client attempts to specify many-to-many relationship at creation.
        '''
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        if serializer.validated_data['assoc_characters'] or serializer.validated_data['assoc_locations']:
            raise NotImplementedError(_('Sorry, but specifying linked characters and locations at creation is not currently supported.'))  # noqa: E501
        serializer = self.perform_create(serializer)
        headers = self.get_success_headers(serializer.data)
        return response.Response(serializer.data, status=status.HTTP_201_CREATED, headers=headers)

    def perform_create(self, serializer):
        '''
        The actual creation part.

        :param serializer:
            The serializer object representing the object to be created.

        :raises ValidationError: if the data submitted is invalid or violates the structure of the tree.

        :returns: A serializer representing the created object.
        '''
        kwarg_dict = {}
        for field in self.fields_required_for_add:
            kwarg_dict[field] = serializer.validated_data[field]
            if self.pos:
                kwarg_dict['pos'] = self.pos
        try:
            with transaction.atomic():
                logger.debug('attempting execution of %s for %s' % (self.action, self.source_node))
                new_node = getattr(self.source_node, self.action)(**kwarg_dict)
                logger.debug("New node created with pk of %s, which is a child of %s" %
                             (new_node.pk, new_node.get_parent(update=True)))
        except ArcGenerationError as AE:
            logger.debug("ArcGenerationError: %s" % str(AE))
            raise ValidationError(str(AE))
        except IntegrityError as IE:
            logger.debug("IntegrityError: %s" % str(IE))
            raise ValidationError(str(IE))
        new_serializer = self.get_serializer_class()(new_node)
        logger.debug("New serializer data looks like: %s" % new_serializer.data)
        return new_serializer


class NodeMoveMixin(object):
    '''
    API mixin for move method for nodes.
    '''
    source_node = None
    target_node = None
    target_node_type_fieldname = None  # fieldname to send to tree_manipulation signal.
    pos = None
    related_key = None  # Specify in subclass

    def get_object(self, pkval):
        '''
        Fetches an individual object and verifies that the user has the appropriate
        permissions to move it.

        :param pkval:
             The primary key value to use in the search.

        :returns: Object

        :raises Http404: if object cannot be found

        :raises PermissionDenied: if user does not have the required permissions.
        '''
        logger.debug('generating queryset')
        queryset = self.filter_queryset(self.get_queryset())
        filter_kwargs = {self.lookup_field: pkval}
        logger.debug('Searching for node with: %s' % filter_kwargs)
        obj = get_object_or_404(queryset, **filter_kwargs)
        self.check_object_permissions(self.request, obj)
        return obj

    def perform_move(self):
        '''
        Validates the proposed move, and if valid, will manipulate the tree accordingly.

        :raises ValidationError: if the move request violates tree structure.

        :raises TreeUnavailable: if the DB tree is out of available nodes.

        :returns: A serializer representing the updated node.
        '''
        logger.debug('Comparing related values from %s' % self.related_key)
        if getattr(self.source_node, self.related_key) != getattr(self.target_node, self.related_key):
            logger.debug('%s and %s do NOT match.' % (getattr(self.source_node, self.related_key),
                                                      getattr(self.target_node, self.related_key)))
            raise ValidationError(_('Nodes must be from the same %s' % self.related_key))
        logger.debug("Keys match!")
        logger.debug("Checking pos vs target...")
        if 'child' not in self.pos and self.target_node.is_root():
            logger.debug("You can't move this to the level of the root!")
            raise ValidationError(_('You cannot move this item to the same level as a root node!'))
        try:
            logger.debug("Sending manipulation signal for %s" % self.source_node.__class__)
            tree_manipulation.send(sender=self.source_node.__class__, instance=self.source_node,
                                   target_node=self.target_node,
                                   target_node_type=getattr(self.target_node, self.target_node_type_fieldname),
                                   action='move', pos=self.pos)
        except IntegrityError as IE:
            logger.error(_('This would result in nodes being children of invalid parent nodes. %s' % str(IE)))
            raise ValidationError(_('This would result in nodes being children of invalid parent nodes. %s' % str(IE)))
        try:
            with transaction.atomic():
                self.source_node.move(target=self.target_node, pos=self.pos)
        except InvalidPosition as IP:  # pragma: no cover Treebeard error that is very hard to trigger.
            logger.error(_("This is not a permitted position. \n Details: %s" % str(IP)))
            raise ValidationError(_('This is not a permitted position. Details: %s' % str(IP)))
        except InvalidMoveToDescendant as IMD:
            logger.error(_('Invalid move to descendant! %s' % str(IMD)))
            raise ValidationError(_('You cannot move an arc element to be the sibling or child of one of its descendants.'))  # noqa: E501
        except PathOverflow as PO:  # pragma: no cover It would take a monumental amount of data to trigger this.
            logger.error(_('There is no further room within the tree for %s!!!!! DETAILS: %s' %
                           (str(self.source_node.__class__), str(PO))))
            raise TreeUnavailable
        return self.get_serializer_class()(self.get_object(self.source_node.pk))

    def post(self, request, *args, **kwargs):
        logger.debug("Trying to fetch source node.")
        self.source_node = self.get_object(kwargs['node_to_move_id'])
        logger.debug("Trying to fetch target node")
        self.target_node = self.get_object(kwargs['target_node_id'])
        self.pos = kwargs['position']
        new_node_serializer = self.perform_move()
        return response.Response(new_node_serializer.data, status=status.HTTP_200_OK)



import logging
from django.db import IntegrityError
from django.utils.translation import gettext_lazy as _
from rest_framework.generics import get_object_or_404
from rest_framework import generics, status, permissions
from rest_framework.response import Response
from rest_framework.exceptions import PermissionDenied, ParseError
from rest_framework_rules.mixins import PermissionRequiredMixin
from fiction_outlines.models import Series, Character, Location, Outline, Arc
from fiction_outlines.models import CharacterInstance, LocationInstance
from fiction_outlines.models import ArcElementNode, StoryElementNode
from fiction_outlines.signals import tree_manipulation
from .serializers import SeriesSerializer, CharacterSerializer, LocationSerializer
from .serializers import OutlineSerializer, ArcSerializer, ArcCreateSerializer, ArcElementNodeSerializer
from .serializers import StoryElementNodeSerializer, CharacterInstanceSerializer, LocationInstanceSerializer
from .mixins import NodeMoveMixin, MultiObjectPermissionsMixin, NodeAddMixin

logger = logging.getLogger('fiction-outlines-api')


class SeriesList(generics.ListCreateAPIView):
    '''
    API view for series list.

    Provides HTTP methods:

    - GET: Retrive :class:`fiction_outlines.models.Series` objects for the current user.
    - POST: Accepts data compatible with a :class:`fiction_outlines_api.serializers.SeriesSerializer`
    '''
    permission_classes = (permissions.IsAuthenticated,)
    serializer_class = SeriesSerializer

    def perform_create(self, serializer):
        serializer.save(user=self.request.user)

    def get_queryset(self):
        return Series.objects.filter(user=self.request.user)


class SeriesDetail(PermissionRequiredMixin, generics.RetrieveUpdateDestroyAPIView):
    '''
    Retrieves details of a series, and enables editing of the object.

    Provides HTTP methods:

    - GET: retrieve the individual object.
    - PUT: Update the object. Takes data compatible with a :class:`fiction_outlines_api.serializers.SeriesSerializer`
    - PATCH: Update the object. Takes partial data compatible with keys from
      :class:`fiction_outlines_api.serializers.SeriesSerializer`
    - DELETE: Delete the object.
    '''
    serializer_class = SeriesSerializer
    object_permission_required = 'fiction_outlines.view_series'
    permission_classes = (permissions.IsAuthenticated,)
    permission_required = 'fiction_outlines_api.valid_user'
    lookup_url_kwarg = 'series'

    def put(self, request, *args, **kwargs):
        self.object_permission_required = 'fiction_outlines.edit_series'
        return super().put(request, *args, **kwargs)

    def delete(self, request, *args, **kwargs):
        self.object_permission_required = 'fiction_outlines.delete_series'
        return super().delete(request, *args, **kwargs)

    def get_queryset(self):
        return Series.objects.all()


class CharacterList(generics.ListCreateAPIView):
    '''
    API view for character list

    Provies HTTP methods:

    - GET: Retrive a list of :class:`fiction_outlines.models.Character` objects for the current user.
    - POST: Create a character. Takes data compatible with
      :class:`fiction_outlines_api.serializers.CharacterSerializer`
    '''
    permission_classes = (permissions.IsAuthenticated,)
    serializer_class = CharacterSerializer

    def perform_create(self, serializer):
        if serializer.validated_data['series']:
            for series in serializer.validated_data['series']:
                if not self.request.user.has_perm('fiction_outlines.edit_series', series):
                    raise PermissionDenied(_('You do not have editing rights for the specified series.'))
        serializer.save(user=self.request.user)

    def get_queryset(self):
        return Character.objects.filter(user=self.request.user)


class CharacterDetail(PermissionRequiredMixin, generics.RetrieveUpdateDestroyAPIView):
    '''
    API view for all single item character operations besides create.

    Provides HTTP methods:

    - GET: Retrieve an individual character record.
    - PUT: Update a character with data compatible with a :class:`fiction_outlines_api.serializers.CharacterSerializer`
    - PATCH: Update a character with partial data that corresponds to keys in
      :class:`fiction_outlines_api.serializers.CharacterSerializer`
    '''
    serializer_class = CharacterSerializer
    object_permission_required = 'fiction_outlines.view_character'
    permission_classes = (permissions.IsAuthenticated,)
    permission_required = 'fiction_outlines_api.valid_user'
    lookup_url_kwarg = 'character'

    def perform_update(self, serializer):
        '''
        Validates that if a series is specified, that the user has rights to it as well.

        :raises PermissionDenied: if the user does not have the required permissions.
        '''
        if 'series' in serializer.validated_data.keys():
            if serializer.validated_data['series']:
                for series in serializer.validated_data['series']:
                    if not self.request.user.has_perm('fiction_outlines.edit_series', series):
                        raise PermissionDenied(_("You do not have editing rights to the specified series."))
        return super().perform_update(serializer)

    def put(self, request, *args, **kwargs):
        self.object_permission_required = 'fiction_outlines.edit_character'
        return super().put(request, *args, **kwargs)

    def patch(self, request, *args, **kwargs):
        self.object_permission_required = 'fiction_outlines.edit_character'
        return super().patch(request, *args, **kwargs)

    def delete(self, request, *args, **kwargs):
        self.object_permission_required = 'fiction_outlines.delete_character'
        return super().delete(request, *args, **kwargs)

    def get_queryset(self):
        return Character.objects.all()


class CharacterInstanceCreateView(MultiObjectPermissionsMixin, PermissionRequiredMixin, generics.CreateAPIView):
    '''
    API view for creating a character instance. Expects kwargs in url for character and
    outline. All other data comes from the serializer.

    Provides HTTP methods:

    - POST: Accepts data compatible with a :class:`fiction_outlines_api.serializers.CharacterInstanceSerializer`
    '''
    serializer_class = CharacterInstanceSerializer
    permission_required = 'fiction_outlines_api.valid_user'
    object_class_permission_dict = {
        'character': {'obj_class': Character, 'object_permission_required': 'fiction_outlines.edit_character',
                      'lookup_url_kwarg': 'character'},
        'outline': {'obj_class': Outline, 'object_permission_required': 'fiction_outlines.edit_outline',
                    'lookup_url_kwarg': 'outline'},
    }

    def perform_create(self, serializer):
        '''
        Attempts the actual creation of the instance.

        :raises ParseError: if the character is already associated with the outline.
        '''
        try:
            serializer.save(character=self.permission_object_dict['character'],
                            outline=self.permission_object_dict['outline'])
        except IntegrityError:
            logger.error("Attempt to create a character instance for a character already present in outline.")
            raise ParseError


class CharacterInstanceDetailView(PermissionRequiredMixin, generics.RetrieveUpdateDestroyAPIView):
    '''
    API view for non-creation actions on CharacterInstance objects.

    Provides HTTP methods:

    - GET: Retrieve an individual object.
    - PUT: Update the object with data compatible with a
      :class:`fiction_outlines_api.serializers.CharacterInstanceSerializer`
    - PATCH: Update the object with partial data that corresponds to keys in the serializer.
    - DELETE: Delete the object.
    '''
    serializer_class = CharacterInstanceSerializer
    object_permission_required = 'fiction_outlines.view_character_instance'
    permission_required = 'fiction_outlines_api.valid_user'
    lookup_url_kwarg = 'instance'

    def put(self, request, *args, **kwargs):
        self.object_permission_required = 'fiction_outlines.edit_character_instance'
        return super().put(request, *args, **kwargs)

    def patch(self, request, *args, **kwargs):
        self.object_permission_required = 'fiction_outlines.edit_character_instance'
        return super().patch(request, *args, **kwargs)

    def delete(self, request, *args, **kwargs):
        self.object_permission_required = 'fiction_outlines.delete_character_instance'
        return super().delete(request, *args, **kwargs)

    def get_queryset(self):
        return CharacterInstance.objects.all()


class LocationList(generics.ListCreateAPIView):
    '''
    API view for location list

    Provides HTTP methods:

    - GET: Retrieve a list of Locations for the current user.
    - POST: Create a location with data compatibile with :class:`fiction_outlines_api.serializers.LocationSerializer`
    '''
    permission_classes = (permissions.IsAuthenticated,)
    serializer_class = LocationSerializer

    def perform_create(self, serializer):
        '''
        Creates the object. If a series is specified, verify the user has rights to it as well.

        :raises PermissionDenied: if the user lackst he required permissions.
        '''
        if serializer.validated_data['series']:
            for series in serializer.validated_data['series']:
                if not self.request.user.has_perm('fiction_outlines.edit_series', series):
                    raise PermissionDenied(_('You do not have editing rights for the specified series.'))
        serializer.save(user=self.request.user)

    def get_queryset(self):
        return Location.objects.filter(user=self.request.user)


class LocationDetail(PermissionRequiredMixin, generics.RetrieveUpdateDestroyAPIView):
    '''
    API view for all single item location operations besides create.

    Provides HTTP methods:

    - GET: Retrieve an individual object.
    - PUT: Update the object with data compatible with a :class:`fiction_outlines_api.serializers.LocationSerializer`
    - PATCH: Update the object with partial data that corresponds to keys in the serializer.
    - DELETE: Delete the object.
    '''
    serializer_class = LocationSerializer
    object_permission_required = 'fiction_outlines.view_location'
    permission_classes = (permissions.IsAuthenticated,)
    permission_required = 'fiction_outlines_api.valid_user'
    lookup_url_kwarg = 'location'

    def perform_update(self, serializer):
        '''
        Performs update actions. If a series is specified, it verifies the user has rights to it as well.

        :raises PermissionDenied: if the user does not have the correct permissions.
        '''
        if 'series' in serializer.validated_data.keys():
            if serializer.validated_data['series']:
                for series in serializer.validated_data['series']:
                    if not self.request.user.has_perm('fiction_outlines.edit_series', series):
                        raise PermissionDenied(_("You do not have editing rights to the specified series."))
        return super().perform_update(serializer)

    def put(self, request, *args, **kwargs):
        self.object_permission_required = 'fiction_outlines.edit_location'
        return super().put(request, *args, **kwargs)

    def patch(self, request, *args, **kwargs):
        self.object_permission_required = 'fiction_outlines.edit_location'
        return super().patch(request, *args, **kwargs)

    def delete(self, request, *args, **kwargs):
        self.object_permission_required = 'fiction_outlines.delete_location'
        return super().delete(request, *args, **kwargs)

    def get_queryset(self):
        return Location.objects.all()


class LocationInstanceCreateView(MultiObjectPermissionsMixin, PermissionRequiredMixin, generics.CreateAPIView):
    '''
    API view for creating a location instance. Expects kwargs in url for location and
    outline. All other data comes from the serializer.

    Provides HTTP method:

    - POST: accepts data compatible with :class:`fiction_outlines_api.serializers.LocationInstanceSerializer`
    '''
    serializer_class = LocationInstanceSerializer
    permission_required = 'fiction_outlines_api.valid_user'
    object_class_permission_dict = {
        'location': {'obj_class': Location, 'object_permission_required': 'fiction_outlines.edit_location',
                     'lookup_url_kwarg': 'location'},
        'outline': {'obj_class': Outline, 'object_permission_required': 'fiction_outlines.edit_outline',
                    'lookup_url_kwarg': 'outline'},
    }

    def perform_create(self, serializer):
        '''
        Creates the location instance.

        :raises ParseError: if the location is already associated with the outline.
        '''
        try:
            serializer.save(location=self.permission_object_dict['location'],
                            outline=self.permission_object_dict['outline'])
        except IntegrityError:
            logger.error("Attempt to create a location instance for a location already present in outline.")
            raise ParseError


class LocationInstanceDetailView(PermissionRequiredMixin, generics.RetrieveDestroyAPIView):
    '''
    API view for non-creation actions on LocationInstance objects. As locations instances don't have
    editable data, only retrieval and destroy are supported at this time.

    Provides HTTP methods:

    - GET: Retrieve an individual object.
    - DELETE: Delete the object.
    '''
    serializer_class = LocationInstanceSerializer
    object_permission_required = 'fiction_outlines.view_location_instance'
    permission_required = 'fiction_outlines_api.valid_user'
    lookup_url_kwarg = 'instance'

    def delete(self, request, *args, **kwargs):
        self.object_permission_required = 'fiction_outlines.delete_location_instance'
        return super().delete(request, *args, **kwargs)

    def get_queryset(self):
        return LocationInstance.objects.all()


class OutlineList(generics.ListCreateAPIView):
    '''
    API view for Outline list

    Provides HTTP methods:

    - GET: Retrived a list of outlines for the current user.
    - POST: Create a new Outline. Accepts data compatible with
      :class:`fiction_outlines_api.serializers.OutlineSerializer`
    '''
    permission_classes = (permissions.IsAuthenticated,)
    serializer_class = OutlineSerializer

    def perform_create(self, serializer):
        '''
        Creates the object. If a series is specified, it verfies that the user has rights to it as well.

        :raises PermissionDenied: if the user does not have the required permissions.
        '''
        if serializer.validated_data['series']:
            if not self.request.user.has_perm('fiction_outlines.edit_series', serializer.validated_data['series']):
                raise PermissionDenied(_('You do not have editing rights for the specified series.'))
        serializer.save(user=self.request.user)

    def get_queryset(self):
        return Outline.objects.filter(
            user=self.request.user
        ).select_related('series').prefetch_related('tags',
                                                    'arc_set',
                                                    'storyelementnode_set',
                                                    'characterinstance_set',
                                                    'locationinstance_set')


class OutlineDetail(PermissionRequiredMixin, generics.RetrieveUpdateDestroyAPIView):
    '''
    API view for all single item outline operations besides create.

    Provides HTTP methods:

    - GET: Retrieve an individual object.
    - PUT: Update the object with data compatible with a :class:`fiction_outlines_api.serializers.OutlineSerializer`
    - PATCH: Update the object with partial data that corresponds to keys in the serializer.
    - DELETE: Delete the object.
    '''
    serializer_class = OutlineSerializer
    object_permission_required = 'fiction_outlines.view_outline'
    permission_classes = (permissions.IsAuthenticated,)
    permission_required = 'fiction_outlines_api.valid_user'
    lookup_url_kwarg = 'outline'

    def perform_update(self, serializer):
        '''
        Updates the object. If a series is specified, it verifies the user has permissions for that object as well.

        :raises PermissionDenied: if the user lacks the needed permissions.
        '''
        if 'series' in serializer.validated_data.keys():
            if (serializer.validated_data['series'] and
                not self.request.user.has_perm('fiction_outlines.edit_series', serializer.validated_data['series'])):
                raise PermissionDenied(_("You do not have editing rights to the specified series."))
        return super().perform_update(serializer)

    def put(self, request, *args, **kwargs):
        self.object_permission_required = 'fiction_outlines.edit_outline'
        return super().put(request, *args, **kwargs)

    def patch(self, request, *args, **kwargs):
        self.object_permission_required = 'fiction_outlines.edit_outline'
        return super().patch(request, *args, **kwargs)

    def delete(self, request, *args, **kwargs):
        self.object_permission_required = 'fiction_outlines.delete_outline'
        return super().delete(request, *args, **kwargs)

    def get_queryset(self):
        return Outline.objects.all().select_related('series').prefetch_related('tags',
                                                                               'arc_set',
                                                                               'storyelementnode_set',
                                                                               'characterinstance_set',
                                                                               'locationinstance_set')


class ArcCreateView(PermissionRequiredMixin, generics.CreateAPIView):
    '''
    API for creating arcs. Uses a custom serializer, as Arcs are generated via special methods inside
    a transaction.

    Provides HTTP method:

    - POST: Accepts data compatible with :class:`fiction_outlines_api.serializers.ArcCreateSerializer`
    '''
    serializer_class = ArcCreateSerializer
    object_permission_required = 'fiction_outlines.edit_outline'
    permission_required = 'fiction_outlines_api.valid_user'

    def dispatch(self, request, *args, **kwargs):
        self.outline = get_object_or_404(Outline, pk=kwargs['outline'])
        return super().dispatch(request, *args, **kwargs)

    def post(self, request, *args, **kwargs):
        self.check_object_permissions(request, self.outline)
        return super().post(request, *args, **kwargs)

    def perform_create(self, serializer):
        self.object = self.outline.create_arc(mace_type=serializer.validated_data['mace_type'],
                                              name=serializer.validated_data['name'])

    def create(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        self.perform_create(serializer)
        result_serializer = ArcSerializer(self.object)
        headers = self.get_success_headers(result_serializer.data)
        return Response(result_serializer.data, status=status.HTTP_201_CREATED, headers=headers)


class ArcDetailView(PermissionRequiredMixin, generics.RetrieveUpdateDestroyAPIView):
    '''
    API for non-create object operations for Arc model.

    Provides HTTP methods:

    - GET: Retrieve an individual object.
    - PUT: Update the object with data compatible with a :class:`fiction_outlines_api.serializers.ArcSerializer`
    - PATCH: Update the object with partial data that corresponds to keys in the serializer.
    - DELETE: Delete the object.
    '''
    serializer_class = ArcSerializer
    object_permission_required = 'fiction_outlines.view_arc'
    permission_required = 'fiction_outlines_api.valid_user'
    lookup_url_kwarg = 'arc'

    def put(self, request, *args, **kwargs):
        self.object_permission_required = 'fiction_outlines.edit_arc'
        return super().put(request, *args, **kwargs)

    def patch(self, request, *args, **kwargs):
        self.object_permission_required = 'fiction_outlines.edit_arc'
        return super().patch(request, *args, **kwargs)

    def delete(self, request, *args, **kwargs):
        self.object_permission_required = 'fiction_outlines.delete_arc'
        return super().delete(request, *args, **kwargs)

    def get_queryset(self):
        return Arc.objects.all()


class ArcNodeDetailView(PermissionRequiredMixin, generics.RetrieveUpdateDestroyAPIView):
    '''
    API for viewing a tree of :class:`fiction_outlines.models.ArcElementNode`, as well as some basic updates.

    Provides HTTP methods:

    - GET: Retrieve an individual object.
    - PUT: Update the object with data compatible with a
      :class:`fiction_outlines_api.serializers.ArcElementNodeSerializer`
    - PATCH: Update the object with partial data that corresponds to keys in the serializer.
    - DELETE: Delete the object.

    '''
    permission_classes = (permissions.IsAuthenticated,)
    serializer_class = ArcElementNodeSerializer
    permission_required = 'fiction_outlines_api.valid_user'
    object_permission_required = 'fiction_outlines.view_arc_node'
    lookup_url_kwarg = 'arcnode'

    def update(self, request, *args, **kwargs):
        try:
            response = super().update(request, *args, **kwargs)
        except IntegrityError as IE:
            response = Response({'error': str(IE)}, status=status.HTTP_400_BAD_REQUEST)
        return response

    def put(self, request, *args, **kwargs):
        self.object_permission_required = 'fiction_outlines.edit_arc_node'
        return super().put(request, *args, **kwargs)

    def patch(self, request, *args, **kwargs):
        self.object_permission_required = 'fiction_outlines.edit_arc_node'
        return super().patch(request, *args, **kwargs)

    def delete(self, request, *args, **kwargs):
        self.object_permission_required = 'fiction_outlines.delete_arc_node'
        return super().delete(request, *args, **kwargs)

    def destroy(self, request, *args, **kwargs):
        '''
        Destroys the object, but first checks that the element doesn't represent either the
        Hook or Resolution of an Arc. These types may not be deleted unless you are deleting the
        whole arc.

        :raises ParseError: if the object represents the Hook or Resolution.
        '''
        instance = self.get_object()
        if instance.arc_element_type in ['mile_hook', 'mile_reso']:
            raise ParseError('You cannot delete the hook or resolution of an arc.')
        self.perform_destroy(instance)
        return Response(status=status.HTTP_204_NO_CONTENT)

    def get_queryset(self):
        return ArcElementNode.objects.all()  # pragma: no cover


class ArcNodeCreateView(PermissionRequiredMixin, NodeAddMixin, generics.CreateAPIView):
    '''
    API view for add_child and add_sibling. You can only create tree objects in relation to
    another object in the tree. See :class:`mixins.NodeAddMixin` for more details.
    '''
    serializer_class = ArcElementNodeSerializer
    permission_required = 'fiction_outlines_api.valid_user'
    object_permission_required = 'fiction_outlines.edit_arc_node'
    fields_required_for_add = ('arc_element_type', 'description', 'story_element_node')
    lookup_url_kwarg = 'arcnode'

    def get_queryset(self):
        return ArcElementNode.objects.all()


class ArcNodeMoveView(PermissionRequiredMixin, NodeMoveMixin, generics.GenericAPIView):
    '''
    View for moving arc nodes. See :class:`mixins.NodeMoveMixin` for more details.
    '''
    permission_classes = (permissions.IsAuthenticated,)
    serializer_class = ArcElementNodeSerializer
    permission_required = 'fiction_outlines_api.valid_user'
    object_permission_required = 'fiction_outlines.edit_arc_node'
    target_node_type_fieldname = 'arc_element_type'
    related_key = 'arc'

    def get_queryset(self):
        return ArcElementNode.objects.all()


class StoryNodeMoveView(PermissionRequiredMixin, NodeMoveMixin, generics.GenericAPIView):
    '''
    View for moving story nodes. See :class:`mixins.NodeMoveMixin` for more details.
    '''
    serializer_class = StoryElementNodeSerializer
    permission_required = 'fiction_outlines_api.valid_user'
    object_permission_required = 'fiction_outlines.edit_story_node'
    target_node_type_fieldname = 'story_element_type'
    related_key = 'outline'

    def get_queryset(self):
        return StoryElementNode.objects.all()


class StoryNodeCreateView(PermissionRequiredMixin, NodeAddMixin, generics.CreateAPIView):
    '''
    View for adding story nodes. You can only create tree elements in relation to other objects in
    the same tree. See :class:`mixins.NodeAddMixin` for more details.
    '''
    serializer_class = StoryElementNodeSerializer
    permission_required = 'fiction_outlines_api.valid_user'
    object_permission_required = 'fiction_outlines.edit_story_node'
    fields_required_for_add = ('story_element_type', 'name', 'description')
    lookup_url_kwarg = 'storynode'

    def get_queryset(self):
        return StoryElementNode.objects.all()


class StoryNodeDetailView(PermissionRequiredMixin, generics.RetrieveUpdateDestroyAPIView):
    '''
    API for viewing and editing a story node.

    Provides HTTP methods:

    - GET: Retrieve an individual object.
    - PUT: Update the object with data compatible with a
      :class:`fiction_outlines_api.serializers.StoryElementNodeSerializer`
    - PATCH: Update the object with partial data that corresponds to keys in the serializer.
    - DELETE: Delete the object.
    '''
    permission_classes = (permissions.IsAuthenticated,)
    serializer_class = StoryElementNodeSerializer
    permission_required = 'fiction_outlines_api.valid_user'
    object_permission_required = 'fiction_outlines.view_story_node'
    lookup_url_kwarg = 'storynode'

    def update(self, request, *args, **kwargs):
        '''
        Updates the node, after first running it through validation to ensure it won't violate the tree
        structure.

        :raises ParseError: if the edit would violate the tree structure.
        '''
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        if 'story_element_type' in serializer.validated_data.keys():
            try:
                instance = self.get_object()
                tree_manipulation.send(
                    sender=StoryElementNode,
                    instance=instance,
                    action='update',
                    target_node_type=serializer.validated_data['story_element_type']
                )
            except IntegrityError as IE:
                logger.debug('Changing the node to this type would result in an invalid outline sructure. Details: %s' % str(IE))  # noqa: E501
                raise ParseError(_(str(IE)))
        try:
            response = super().update(request, *args, **kwargs)
        except IntegrityError as IE:
            response = Response({'error': str(IE)}, status=status.HTTP_400_BAD_REQUEST)
        return response

    def put(self, request, *args, **kwargs):
        self.object_permission_required = 'fiction_outlines.edit_story_node'
        return super().put(request, *args, **kwargs)

    def patch(self, request, *args, **kwargs):
        self.object_permission_required = 'fiction_outlines.edit_story_node'
        return super().patch(request, *args, **kwargs)

    def delete(self, request, *args, **kwargs):
        self.object_permission_required = 'fiction_outlines.delete_story_node'
        return super().delete(request, *args, **kwargs)

    def get_queryset(self):
        return StoryElementNode.objects.all()  # pragma: no cover
